# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Update assets to prepare for release."""

import argparse
import shutil
import tempfile
from collections import Counter
from git import Repo
from pathlib import Path
from typing import List

import azureml.assets as assets
import azureml.assets.environment as environment
import azureml.assets.util as util
from azureml.assets.util import logger

ASSET_COUNT = "asset_count"
UPDATED_COUNT = "updated_count"
UPDATED_ENV_COUNT = "updated_env_count"
COUNTERS = [ASSET_COUNT, UPDATED_COUNT, UPDATED_ENV_COUNT]


def pin_env_files(env_config: assets.EnvironmentConfig):
    """Pin versions in environment files marked as templates.

    Args:
        env_config (assets.EnvironmentConfig): Environment config

    Raises:
        Exception: Any failures to pin
    """
    files_to_pin = env_config.template_files_with_path

    # Replace template tags in environment config
    if assets.Config._contains_template(env_config.image_name):
        files_to_pin.append(env_config.file_name_with_path)

    # Replace template tags in files to pin
    for file_to_pin in files_to_pin:
        if file_to_pin.exists():
            try:
                environment.transform_file(file_to_pin)
            except Exception as e:
                raise Exception(f"Failed to pin versions in {file_to_pin}: {e}")
        else:
            logger.log_warning(f"Failed to pin versions in {file_to_pin}: File not found")


def release_tag_exists(asset_config: assets.AssetConfig, release_directory_root: Path) -> bool:
    """Check repo for an asset's release tag.

    Args:
        asset_config (assets.AssetConfig): Asset config
        release_directory_root (Path): Release branch location

    Returns:
        bool: True if the tag exists, False otherwise
    """
    # Check git repo for version-specific tag
    repo = Repo(release_directory_root)
    return asset_config.full_name in repo.tags


def get_latest_release_tag_version(asset_config: assets.AssetConfig, release_directory_root: Path) -> str:
    """Check repo to see if an asset's previous version was released if a latest tag exists.

    Args:
        asset_config (assets.AssetConfig): Asset config
        release_directory_root (Path): Release branch location

    Returns:
        str: Latest version found, or None if asset is not an environment or no tags not found
    """
    repo = Repo(release_directory_root)
    tags = [t for t in repo.tags if t.name.startswith(f"{asset_config.partial_name}/")]

    if not tags:
        # No releases
        return None

    # Get the latest tag
    ordered_tags = sorted(tags, key=lambda t: t.commit.authored_datetime)
    latest_tag = ordered_tags[-1].name
    _, _, latest_version = assets.AssetConfig.parse_full_name(latest_tag)

    return latest_version


def _update_asset_files(asset_config: assets.AssetConfig):
    """Update asset files.

    Args:
        asset_config (assets.AssetConfig): Asset config
    """
    if asset_config.type == assets.AssetType.ENVIRONMENT:
        env_config = asset_config.extra_config_as_object()
        pin_env_files(env_config)


def update_asset(asset_config: assets.AssetConfig,
                 output_directory_root: Path = None,
                 release_directory_root: Path = None,
                 skip_unreleased: bool = False,
                 use_version_dir: bool = False) -> str:
    """Update asset to prepare for release.

    Args:
        asset_config (assets.AssetConfig): Asset config
        output_directory_root (Path, optional): Output directory for updated assets. Defaults to None.
        release_directory_root (Path, optional): Release branch location. Defaults to None.
        skip_unreleased (bool, optional): Skip unreleased explicitly-versioned assets. Defaults to False.
        use_version_dir (bool, optional): Use version directory for output. Defaults to False.

    Returns:
        str: Version of updated asset, or None if not updated
    """
    # The release directory is required if auto-versioning
    if asset_config.auto_version and release_directory_root is None:
        logger.log_error(f"Asset {asset_config.name} is auto-versioned but can't be updated because no release "
                         "directory was specified to compare against")
        exit(1)

    # To keep things simple, we'll create a temporary directory for each update
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)

        # Copy asset to temp directory and pin image/package versions
        if output_directory_root is not None or release_directory_root is not None:
            temp_asset_dir = util.copy_asset_to_output_dir(asset_config=asset_config, output_directory=temp_dir_path,
                                                           add_subdir=True)
            temp_asset_config = util.find_assets(input_dirs=temp_dir_path,
                                                 asset_config_filename=asset_config.file_name)[0]
            _update_asset_files(temp_asset_config)

        # Get version info, set a few defaults
        main_version = asset_config.version
        auto_version = asset_config.auto_version
        release_version = None
        pending_release = False

        # Look in release directory for existing asset
        if release_directory_root is not None:
            release_dir = util.get_asset_release_dir(asset_config, release_directory_root)
            if release_dir.exists():
                # Check existing release dir
                release_asset_configs = util.find_assets(input_dirs=release_dir,
                                                         asset_config_filename=asset_config.file_name)
                if not release_asset_configs:
                    logger.log_error(f"Release directory {release_dir} exists, but it's missing an asset config file")
                    exit(1)
                release_asset_config = release_asset_configs[0]
                release_version = release_asset_config.version

                # Update spec file
                assets.update_spec(temp_asset_config, version=release_version)

                # Compare temporary version with one in release
                dirs_equal = util.are_dir_trees_equal(temp_asset_dir, release_dir)
                if dirs_equal:
                    return None

                # See if the asset version is unreleased
                pending_release = not release_tag_exists(release_asset_config, release_directory_root)
                if pending_release and not auto_version and main_version != release_version and skip_unreleased:
                    # Skip the unreleased asset version
                    logger.log_warning(f"Skipping {release_asset_config.type.value} {release_asset_config.name} "
                                       f"because version {release_version} hasn't been released yet")
                    return None

        # Determine new version
        if not auto_version:
            # Use explicit version
            new_version = main_version
        elif pending_release:
            # Reuse existing auto version
            new_version = release_version
        else:
            # Increment auto version
            new_version = str(int(release_version) + 1 if release_version else 1)

        # Just update in place if no output directory
        if output_directory_root is None:
            _update_asset_files(asset_config)
            assets.update_spec(asset_config, version=new_version)
            return new_version

        # Identify asset's output directory
        output_is_release = (release_directory_root is not None and output_directory_root is not None
                             and output_directory_root.exists()
                             and output_directory_root.samefile(release_directory_root))
        if output_is_release:
            # Prevent release directory corruption
            use_version_dir = False
        output_directory = util.get_asset_output_dir(asset_config, output_directory_root, use_version_dir)

        # Copy and replace any existing directory
        util.copy_replace_dir(source=temp_asset_dir, dest=output_directory)

        # Update version in spec by copying clean spec and updating it
        asset_config_relative_path = temp_asset_config.file_name_with_path.relative_to(temp_asset_dir)
        output_asset_config = assets.AssetConfig(output_directory / asset_config_relative_path)
        shutil.copyfile(asset_config.spec_with_path, output_asset_config.spec_with_path)
        assets.update_spec(output_asset_config, version=new_version)

        return new_version


def update_assets(input_dirs: List[Path],
                  asset_config_filename: str,
                  output_directory_root: Path = None,
                  release_directory_root: Path = None,
                  skip_unreleased: bool = False,
                  use_version_dirs: bool = False):
    """Update assets to prepare for release.

    Args:
        input_dirs (List[Path]): List of directories to search for assets.
        asset_config_filename (str): Asset config filename to search for
        output_directory_root (Path, optional): Output directory for updated assets. Defaults to None.
        release_directory_root (Path, optional): Release directory location. Defaults to None.
        skip_unreleased (bool, optional): Skip unreleased explicitly-versioned assets. Defaults to False.
        use_version_dirs (bool, optional): Use version directories for output. Defaults to False.
    """
    # Find assets under input dirs
    counters = Counter()
    for asset_config in util.find_assets(input_dirs, asset_config_filename):
        counters[ASSET_COUNT] += 1

        # Update asset if it's changed
        new_version = update_asset(asset_config=asset_config,
                                   output_directory_root=output_directory_root,
                                   release_directory_root=release_directory_root,
                                   skip_unreleased=skip_unreleased,
                                   use_version_dir=use_version_dirs)
        if new_version:
            logger.print(f"Updated {asset_config.type.value} {asset_config.name} version {new_version}")
            counters[UPDATED_COUNT] += 1

            # Track updated environments
            if asset_config.type == assets.AssetType.ENVIRONMENT:
                counters[UPDATED_ENV_COUNT] += 1
        else:
            logger.log_debug(f"No changes detected for {asset_config.type.value} {asset_config.name}")
    logger.print(f"{counters[UPDATED_COUNT]} of {counters[ASSET_COUNT]} asset(s) updated")

    # Set variables
    for counter_name in COUNTERS:
        logger.set_output(counter_name, counters[counter_name])


if __name__ == '__main__':
    # Handle command-line args
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input-dirs", required=True,
                        help="Comma-separated list of directories containing assets")
    parser.add_argument("-a", "--asset-config-filename", default=assets.DEFAULT_ASSET_FILENAME,
                        help="Asset config file name to search for")
    parser.add_argument("-o", "--output-directory", type=Path,
                        help="Copy new/updated assets into this directory, can be the same as --release-directory "
                        "and if omitted, assets will be updated in place")
    parser.add_argument("-r", "--release-directory", type=Path,
                        help="Directory to which the release branch has been cloned; if specified, only unreleased "
                        "and updated assets will be copied to the output directory")
    parser.add_argument("-s", "--skip-unreleased", action="store_true",
                        help="Skip unreleased explicitly-versioned assets in the release branch")
    parser.add_argument("-v", "--use-version-dirs", action="store_true",
                        help="Use version directories when storing assets in output directory")
    args = parser.parse_args()

    # Check interdependencies
    if args.skip_unreleased and args.release_directory is None:
        parser.error("--skip-unreleased requires --release-directory")
    if args.use_version_dirs and args.output_directory is None:
        parser.error("--use-version-dirs requires --output-directory")

    # Convert comma-separated values to lists
    input_dirs = [Path(d) for d in args.input_dirs.split(",")]

    # Update assets
    update_assets(input_dirs=input_dirs,
                  asset_config_filename=args.asset_config_filename,
                  output_directory_root=args.output_directory,
                  release_directory_root=args.release_directory,
                  skip_unreleased=args.skip_unreleased,
                  use_version_dirs=args.use_version_dirs)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Deployment config classes."""

from dataclasses import dataclass
from typing import Dict, List
import azureml.assets as assets
from marshmallow import Schema, fields, post_load, validates, validates_schema
from ruamel.yaml import YAML

"""
Sample deplyment config YAML file:

create: # Assets to create
  component: # List of components
    - component1
  environment: # List of environments
    - environment1

update: # Assets to update
  components: # Not yet supported
  environment: # Dictionary of environments
    acpt-pytorch-1.11-cuda11.5: # Environment name
      # List containing updates to make, organized by versions to affect
      - versions: ["1", "2"] # Specific versions to update; don't use with all_versions
        all_versions: true # Update all versions; don't use with versions
        description: "DEPRECATED: Please use acpt-pytorch-1.11-cuda11.3 instead." # Replaces any existing description
        tags: # Tag updates
          add: # Add or update existing tags
            Tag1: "value"
            Deprecated: "" # Tags with no values are okay
          replace: # Replace any existing tags with this set; don't use with add or delete
            Tag1: "value"
            Tag2: "value"
          delete: # Tags to delete; delete wins any overlap with add
            - Tag1
            - Tag2
        stage: "Archived" # Use Active or Archived to control visibility to list operations

delete: # Assets to delete
  component: # List of components
    name: microsoft_azureml_automl_classification_component
      - versions: ["2"] # Specific versions to delete; don't use with all_versions
        all_versions: true # Deletes all versions
        delete_container: true # Also delete asset container
"""


@dataclass
class AssetTags:
    """Asset tags class.

    Args:
        add (Dict[str, str]): Tags to add.
        replace (Dict[str, str]): Replace any existing tags with these.
        delete (List[str]): Tags to delete.
    """

    add: Dict[str, str] = None
    replace: Dict[str, str] = None
    delete: List[str] = None


@dataclass
class AssetProperties:
    """Asset properties class.

    Args:
        add (Dict[str, str]): Properties to add.
    """

    add: Dict[str, str] = None


@dataclass
class Versions:
    """Versions base class.

    Args:
        versions (List[str]): Versions to affect.
        all_versions (bool): Affect all versions.
    """

    versions: List[str] = None
    all_versions: bool = False


@dataclass
class AssetVersionUpdate(Versions):
    """Asset version update class.

    Args:
        description (str): New description.
        tags (AssetTags): Tag updates.
        stage (str): New stage.
    """

    description: str = None
    tags: AssetTags = None
    properties: AssetProperties = None
    stage: str = None

    def __post_init__(self):
        """Convert field values to objects."""
        if self.tags:
            self.tags = AssetTags(**self.tags)

        if self.properties:
            self.properties = AssetProperties(**self.properties)


@dataclass
class AssetUpdate:
    """Asset update class.

    Args:
        name (str): Asset name.
        updates (List[AssetVersionUpdate]): Updates to apply.
    """

    name: str = None
    updates: List[AssetVersionUpdate] = None

    def __post_init__(self):
        """Convert field values to objects."""
        if self.updates:
            self.updates = [AssetVersionUpdate(**u) for u in self.updates]


@dataclass
class AssetVersionDelete(Versions):
    """Asset version delete class.

    Args:
        delete_container (bool): Delete container.
    """

    delete_container: bool = False


@dataclass
class AssetDelete:
    """Asset delete class.

    Args:
        name (str): Asset name.
        deletes (List[AssetVersionDelete]): Deletes to apply.
    """

    name: str = None
    deletes: List[AssetVersionDelete] = None

    def __post_init__(self):
        """Convert field values to objects."""
        if self.deletes:
            self.deletes = [AssetVersionDelete(**d) for d in self.deletes]


@dataclass
class DeploymentConfig:
    """Deployment config class.

    Args:
        create (Dict[assets.AssetType, List[str]]): Assets to create.
        update (Dict[assets.AssetType, List[AssetUpdate]]): Assets to update.
        delete (Dict[assets.AssetType, List[AssetDelete]]): Assets to delete.
    """

    create: Dict[assets.AssetType, List[str]] = None
    update: Dict[assets.AssetType, List[AssetUpdate]] = None
    delete: Dict[assets.AssetType, List[AssetDelete]] = None

    def __post_init__(self):
        """Convert field values to objects."""
        if self.update:
            self.update = {k: self._convert_update_dict(v) for k, v in self.update.items()}
        if self.delete:
            self.delete = {k: self._convert_delete_dict(v) for k, v in self.delete.items()}

    @staticmethod
    def _convert_update_dict(update_assets: Dict[str, object]) -> List[AssetUpdate]:
        return [AssetUpdate(name=k, updates=v) for k, v in update_assets.items()]

    @staticmethod
    def _convert_delete_dict(delete_assets: Dict[str, object]) -> List[AssetDelete]:
        return [AssetDelete(name=k, deletes=v) for k, v in delete_assets.items()]

    @staticmethod
    def load(deployment_config: str) -> "DeploymentConfig":
        """Load a deployment config from a file.

        Args:
            deployment_config (str): Deployment config file.

        Returns:
            DeploymentConfig: Deployment config.
        """
        with open(deployment_config) as fp:
            config = YAML().load(fp)
            return DeploymentConfigSchema().load(config)

    def should_create(self, asset_type: assets.AssetType, asset_name: str) -> bool:
        """Determine if an asset should be created.

        Args:
            asset_type (assets.AssetType): Asset type.
            asset_name (str): Asset name.

        Returns:
            bool: True if the asset should be created.
        """
        return any(n in {"*", asset_name} for n in self.create.get(asset_type, []))


class TagsSchema(Schema):
    """Tags schema."""

    add = fields.Dict(fields.Str(), fields.Str())
    replace = fields.Dict(fields.Str(), fields.Str())
    delete = fields.List(fields.Str())

    @validates('add')
    def _validate_add(self, value: Dict[str, str]):
        if value is not None and not value:
            raise ValueError("add must be non-empty")

    @validates('delete')
    def _validate_delete(self, value: List[str]):
        if value is not None and not value:
            raise ValueError("delete must be non-empty")

    @validates_schema
    def _validate_schema(self, data: Dict[str, object], **kwargs):
        if data.get('replace') and (data.get('add') or data.get('delete')):
            raise ValueError("replace can't be used with add or delete")


class PropertiesSchema(Schema):
    """Properties schema."""

    add = fields.Dict(fields.Str(), fields.Str())

    @validates('add')
    def _validate_add(self, value: Dict[str, str]):
        if value is not None and not value:
            raise ValueError("add must be non-empty")


class VersionsSchema(Schema):
    """Versions schema."""

    versions = fields.List(fields.Str())
    all_versions = fields.Boolean()

    @validates('versions')
    def _validate_versions(self, value: List[str]):
        if value is not None and not value:
            raise ValueError("versions must be non-empty")

    @validates_schema
    def _validate_schema(self, data: Dict[str, object], **kwargs):
        if data.get('versions') and data.get('all_versions'):
            raise ValueError("Only one of versions and all_versions can be specified")


class AssetVersionUpdateSchema(VersionsSchema):
    """Asset version update schema."""

    description = fields.Str()
    tags = fields.Nested(TagsSchema)
    properties = fields.Nested(PropertiesSchema)
    stage = fields.Str()


class AssetVersionDeleteSchema(VersionsSchema):
    """Asset version delete schema."""

    delete_container = fields.Boolean()


class DeploymentConfigSchema(Schema):
    """Deployment config schema."""

    create = fields.Dict(fields.Enum(assets.AssetType, by_value=True), fields.List(fields.Str()))
    update = fields.Dict(
        fields.Enum(assets.AssetType, by_value=True),
        fields.Dict(
            fields.Str(),
            fields.List(fields.Nested(AssetVersionUpdateSchema))
        )
    )
    delete = fields.Dict(
        fields.Enum(assets.AssetType, by_value=True),
        fields.Dict(
            fields.Str(),
            fields.List(fields.Nested(AssetVersionDeleteSchema))
        )
    )

    @post_load
    def _convert_to_object(self, data, **kwargs):
        # Convert to objects
        return DeploymentConfig(**data)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Provide unified logging across GitHub and Azure DevOps, with a fallback to console."""

import os
import threading

_stdout_lock = threading.Lock()
_output_lock = threading.Lock()


class Logger:
    """Base logger class."""

    def log_debug(self, message: str, title: str = None):
        """Log a debug message.

        Args:
            message (str): Message.
            title (str, optional): Message title. Defaults to None.
        """
        pass

    def log_warning(self, message: str, title: str = None):
        """Log a warning message.

        Args:
            message (str): Message.
            title (str, optional): Message title. Defaults to None.
        """
        pass

    def log_error(self, message: str, title: str = None):
        """Log an error message.

        Args:
            message (str): Message.
            title (str, optional): Message title. Defaults to None.
        """
        pass

    def start_group(self, title: str):
        """Start a collapsible group.

        Args:
            title (str): Title.
        """
        pass

    def end_group(self):
        """End a collapsible group."""
        pass

    def set_output(self, name: str, value: str):
        """Set an output variable.

        Args:
            name (str): Name.
            value (str): Value.
        """
        pass

    def print(self, message: str):
        """Print a message, blocking until complete.

        Args:
            message (str): Message.
        """
        with _stdout_lock:
            print(message)


class GitHubLogger(Logger):
    """Logger running under a GitHub agent."""

    def log_debug(self, message: str, title: str = None):
        """Log a debug message.

        Args:
            message (str): Message.
            title (str, optional): Message title. Defaults to None.
        """
        self._log("debug", message, title)

    def log_warning(self, message: str, title: str = None):
        """Log a warning message.

        Args:
            message (str): Message.
            title (str, optional): Message title. Defaults to None.
        """
        self._log("warning", message, title)

    def log_error(self, message: str, title: str = None):
        """Log an error message.

        Args:
            message (str): Message.
            title (str, optional): Message title. Defaults to None.
        """
        self._log("error", message, title)

    def start_group(self, title: str):
        """Start a collapsible group.

        Args:
            title (str): Title.
        """
        self.print(f"::group::{title}")

    def end_group(self):
        """End a collapsible group."""
        self.print("::endgroup::")

    def set_output(self, name: str, value: str):
        """Set an output variable.

        Args:
            name (str): Name.
            value (str): Value.
        """
        output_filename = os.environ["GITHUB_OUTPUT"]
        with _output_lock, open(output_filename, "a") as f:
            f.write(f"{name}={value}\n")

    def _log(self, log_level: str, message: str, title: str = None):
        title_string = f" title={title}" if title is not None else ""
        self.print(f"::{log_level}{title_string}::{message}")


class AzureDevOpsLogger(Logger):
    """Logger running under an Azure DevOps agent."""

    def log_debug(self, message: str, title: str = None):
        """Log a debug message.

        Args:
            message (str): Message.
            title (str, optional): Message title. Defaults to None.
        """
        self._log("debug", message, title)

    def log_warning(self, message: str, title: str = None):
        """Log a warning message.

        Args:
            message (str): Message.
            title (str, optional): Message title. Defaults to None.
        """
        self._log("warning", message, title)

    def log_error(self, message: str, title: str = None):
        """Log an error message.

        Args:
            message (str): Message.
            title (str, optional): Message title. Defaults to None.
        """
        self._log("error", message, title)

    def start_group(self, title: str):
        """Start a collapsible group.

        Args:
            title (str): Title.
        """
        self.print(f"##[group]{title}")

    def end_group(self):
        """End a collapsible group."""
        self.print("##[endgroup]")

    def set_output(self, name: str, value: str):
        """Set an output variable.

        Args:
            name (str): Name.
            value (str): Value.
        """
        self.print(f"##vso[task.setvariable variable={name};isoutput=true]{value}")

    def _log(self, log_level: str, message: str, title: str = None):
        if title is not None:
            title_string = f";title={title}" if title is not None else ""
            self.print(f"##vso[task.logissue type={log_level}{title_string}]{message}")
        else:
            self.print(f"##[{log_level}]{message}")


class ConsoleLogger(Logger):
    """Logger running at the console."""

    def log_debug(self, message: str, title: str = None):
        """Log a debug message.

        Args:
            message (str): Message.
            title (str, optional): Ignored by this logger.
        """
        self._log("debug", message, title)

    def log_warning(self, message: str, title: str = None):
        """Log a warning message.

        Args:
            message (str): Message.
            title (str, optional): Message title.  Ignored by this logger.
        """
        self._log("warning", message, title)

    def log_error(self, message: str, title: str = None):
        """Log an error message.

        Args:
            message (str): Message.
            title (str, optional): Message title. Ignored by this logger.
        """
        self._log("error", message, title)

    def start_group(self, title: str):
        """Ignored by this logger."""
        pass

    def end_group(self):
        """Ignored by this logger."""
        pass

    def set_output(self, name: str, value: str):
        """Ignored by this logger."""
        pass

    def _log(self, log_level: str, message: str, title: str = None):
        self.print(f"{log_level}: {message}")


def _create_default_logger() -> Logger:
    if os.environ.get("GITHUB_RUN_NUMBER") is not None:
        return GitHubLogger()
    elif os.environ.get("BUILD_BUILDNUMBER") is not None:
        return AzureDevOpsLogger()
    else:
        return ConsoleLogger()


# Create default logger
logger = _create_default_logger()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Set up azureml-assets package."""

from setuptools import setup, find_packages

setup(
   name="azureml-assets",
   version="1.16.1",
   description="Utilities for publishing assets to Azure Machine Learning system registries.",
   author="Microsoft Corp",
   packages=find_packages(),
   install_requires=[
      "GitPython>=3.1",
      "ruamel.yaml==0.17.21",
      "pip>=21",
      "marshmallow>=3.19",
      "tenacity>=8.2.2",
      "azure-ai-ml>=1.9.0",
   ],
   python_requires=">=3.8,<4.0",
   license="MIT",
   classifiers=[
      "Development Status :: 5 - Production/Stable",
      "Intended Audience :: Developers",
      "Topic :: Software Development :: Build Tools",
      "License :: OSI Approved :: MIT License",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3.11",
   ],
   project_urls={
        # 'Documentation': "https://example.com/documentation/",
        'Source': "https://github.com/Azure/azureml-assets/",
        'Changelog': "https://github.com/Azure/azureml-assets/blob/main/scripts/azureml-assets/CHANGELOG.md",
    },
)
